@@ -38,8 +38,8 @@ class PyKeePass(object):
     """Open a KeePass database
 
     Args:
-        filename (:obj:`str`, optional): path to database.  If None, the
-            path given when the database was opened is used.
+        filename (:obj:`str`, optional): path to database or stream object.
+            If None, the path given when the database was opened is used.
         password (:obj:`str`, optional): database password.  If None,
             database is assumed to have no password
         keyfile (:obj:`str`, optional): path to keyfile.  If None,
@@ -92,12 +92,21 @@ class PyKeePass(object):
             filename = self.filename
 
         try:
-            self.kdbx = KDBX.parse_file(
-                filename,
-                password=password,
-                keyfile=keyfile,
-                transformed_key=transformed_key
-            )
+            if hasattr(filename, "read"):
+                self.kdbx = KDBX.parse_stream(
+                    filename,
+                    password=password,
+                    keyfile=keyfile,
+                    transformed_key=transformed_key
+                )
+            else:
+                self.kdbx = KDBX.parse_file(
+                    filename,
+                    password=password,
+                    keyfile=keyfile,
+                    transformed_key=transformed_key
+                )
+
         except ChecksumError as e:
             if e.path in (
                     '(parsing) -> body -> cred_check', # KDBX4
@@ -114,25 +123,42 @@ class PyKeePass(object):
             else:
                 raise
 
+    def reload(self):
+        """Reload current database using previous credentials """
+
+        self.read(self.filename, self.password, self.keyfile)
+
     def save(self, filename=None, transformed_key=None):
         """Save current database object to disk.
 
         Args:
-            filename (:obj:`str`, optional): path to database.  If None, the
-                path given when the database was opened is used.
+            filename (:obj:`str`, optional): path to database or stream object.
+                If None, the path given when the database was opened is used.
+                PyKeePass.filename is unchanged.
             transformed_key (:obj:`bytes`, optional): precomputed transformed
                 key.
         """
+        output = None
         if not filename:
             filename = self.filename
 
-        KDBX.build_file(
-            self.kdbx,
-            filename,
-            password=self.password,
-            keyfile=self.keyfile,
-            transformed_key=transformed_key
-        )
+        if hasattr(filename, "write"):
+            output = KDBX.build_stream(
+                self.kdbx,
+                filename,
+                password=self.password,
+                keyfile=self.keyfile,
+                transformed_key=transformed_key
+            )
+        else:
+            output = KDBX.build_file(
+                self.kdbx,
+                filename,
+                password=self.password,
+                keyfile=self.keyfile,
+                transformed_key=transformed_key
+            )
+        return output
 
     @property
     def version(self):
@@ -178,6 +204,13 @@ class PyKeePass(object):
         """Group: root Group of database"""
         return self.find_groups(path='', first=True)
 
+    @property
+    def recyclebin_group(self):
+        """Group: RecycleBin Group of database"""
+        elem = self._xpath('/KeePassFile/Meta/RecycleBinUUID', first=True)
+        recyclebin_uuid = uuid.UUID( bytes = base64.b64decode(elem.text) )
+        return self.find_groups(uuid=recyclebin_uuid, first=True)
+
     @property
     def groups(self):
         """:obj:`list` of :obj:`Group`: list of all Group objects in database
@@ -244,7 +277,7 @@ class PyKeePass(object):
 
         if tree is None:
             tree = self.tree
-        logger.debug(xpath_str)
+        logger.debug('xpath query: ' + xpath_str)
         elements = tree.xpath(
             xpath_str, namespaces={'re': 'http://exslt.org/regular-expressions'}
         )
@@ -280,14 +313,12 @@ class PyKeePass(object):
             first = True
 
             xp += '/KeePassFile/Root/Group'
-            path = path.strip('/')
             # split provided path into group and element
-            group_path = os.path.dirname(path)
-            element = os.path.basename(path)
+            group_path = path[:-1]
+            element = path[-1] if len(path) > 0 else ''
             # build xpath from group_path and element
-            if group_path:
-                for group in group_path.split('/'):
-                    xp += path_xp[regex]['group'].format(group, flags=flags)
+            for group in group_path:
+                xp += path_xp[regex]['group'].format(group, flags=flags)
             if 'Entry' in prefix:
                 xp += path_xp[regex]['entry'].format(element, flags=flags)
             elif element and 'Group' in prefix:
@@ -333,6 +364,17 @@ class PyKeePass(object):
 
         return res
 
+    def _can_be_moved_to_recyclebin(self, entry_or_group):
+        if entry_or_group == self.root_group:
+            return False
+        recyclebin_group = self.recyclebin_group
+        if recyclebin_group is None:
+            return True
+        uuid_str = base64.b64encode( entry_or_group.uuid.bytes).decode('utf-8')
+        elem = self._xpath('./UUID[text()="{}"]/..'.format(uuid_str), tree=recyclebin_group._element, first=True, history=False, cast=False)
+        return elem is None
+
+
     # ---------- Groups ----------
 
     def find_groups(self, recursive=True, path=None, group=None, **kwargs):
@@ -424,6 +466,40 @@ class PyKeePass(object):
     def move_group(self, group, destination_group):
         destination_group.append(group)
 
+    def _create_or_get_recyclebin_group(self, **kwargs):
+        existing_group = self.recyclebin_group
+        if existing_group is not None:
+            return existing_group
+        kwargs.setdefault('group_name', 'Recycle Bin')
+        group = self.add_group( self.root_group, **kwargs)
+        elem = self._xpath('/KeePassFile/Meta/RecycleBinUUID', first=True)
+        elem.text = base64.b64encode(group.uuid.bytes).decode('utf-8')
+        return group
+
+    def trash_group(self, group):
+        """Move a group to the RecycleBin
+
+        Args:
+            group (:obj:`Group`): Group to send to the RecycleBin
+        """
+        if not self._can_be_moved_to_recyclebin(group):
+            raise UnableToSendToRecycleBin
+        recyclebin_group = self._create_or_get_recyclebin_group()
+        self.move_group( group, recyclebin_group)
+
+    def empty_group(self, group):
+        """Delete the content of a group.
+
+        This does not delete the group itself
+
+        Args:
+            group (:obj:`Group`): Group to empty
+        """
+        while len(group.subgroups):
+            self.delete_group(group.subgroups[0])
+        while len(group.entries):
+            self.delete_entry(group.entries[0])
+
     # ---------- Entries ----------
 
     def find_entries(self, recursive=True, path=None, group=None, **kwargs):
@@ -563,6 +639,17 @@ class PyKeePass(object):
     def move_entry(self, entry, destination_group):
         destination_group.append(entry)
 
+    def trash_entry(self, entry):
+        """Move an entry to the RecycleBin
+
+        Args:
+            entry (:obj:`Entry`): Entry to send to the RecycleBin
+        """
+        if not self._can_be_moved_to_recyclebin(entry):
+            raise UnableToSendToRecycleBin
+        recyclebin_group = self._create_or_get_recyclebin_group()
+        self.move_entry( entry, recyclebin_group)
+
     # ---------- Attachments ----------
 
     def find_attachments(self, recursive=True, path=None, element=None, **kwargs):
@@ -657,13 +744,16 @@ class PyKeePass(object):
             reference.id = reference.id - 1
 
 
-def create_database(filename, password=None, keyfile=None, transformed_key=None):
-    keepass_instance = PyKeePass(BLANK_DATABASE_LOCATION, BLANK_DATABASE_PASSWORD)
+def create_database(
+        filename, password=None, keyfile=None, transformed_key=None
+):
+    keepass_instance = PyKeePass(
+        BLANK_DATABASE_LOCATION, BLANK_DATABASE_PASSWORD
+    )
 
     keepass_instance.filename = filename
     keepass_instance.password = password
     keepass_instance.keyfile = keyfile
 
     keepass_instance.save(transformed_key)
-
     return keepass_instance
